@@ -7,6 +7,7 @@ from streprogen.exercises import DynamicExercise, StaticExercise
 from streprogen.mealplan import Mealplan
 from streprogen.modeling import (
     progression_diffeq,
+    progression_sinh,
     progression_sinusoidal,
     progression_sawtooth,
     reps_to_intensity,
@@ -16,7 +17,7 @@ from streprogen.modeling import (
 from streprogen.optimization import RepSchemeGenerator, RepSchemeOptimizer
 from streprogen.program import Program
 
-__version__ = "3.1.0"
+__version__ = "3.3.0"
 
 __all__ = [
     "StaticExercise",
@@ -27,6 +28,7 @@ __all__ = [
     "reps_to_intensity_tight",
     "reps_to_intensity_relaxed",
     "progression_diffeq",
+    "progression_sinh",
     "progression_sinusoidal",
     "progression_sawtooth",
     "Food",
